@@ -4,16 +4,20 @@
 /*
  * conffile.h	Defines for the conffile parsing routines.
  *
- * Version:	$Id$
+ * Version:	$Id: 21fc2cd5e44ab325949322416a674d50d05ea259 $
  *
  */
 
 #include <freeradius-devel/ident.h>
-RCSIDH(conffile_h, "$Id$")
+RCSIDH(conffile_h, "$Id: 21fc2cd5e44ab325949322416a674d50d05ea259 $")
 
 #include <stddef.h>
 #include <freeradius-devel/token.h>
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /*
  * Export the minimum amount of information about these structs
  */
@@ -54,12 +58,15 @@ int		cf_section_parse(CONF_SECTION *, void *base,
 				 const CONF_PARSER *variables);
 void		cf_section_parse_free(CONF_SECTION *cs, void *base);
 const CONF_PARSER *cf_section_parse_table(CONF_SECTION *cs);
+int cf_exclude_file(const char *filename);
 CONF_SECTION	*cf_file_read(const char *file);
 int		cf_file_include(const char *file, CONF_SECTION *cs);
 
 CONF_PAIR	*cf_pair_find(const CONF_SECTION *, const char *name);
 CONF_PAIR	*cf_pair_find_next(const CONF_SECTION *, CONF_PAIR *, const char *name);
 CONF_SECTION	*cf_section_find(const char *name);
+CONF_SECTION	*cf_section_find_name2(const CONF_SECTION *section,
+				       const char *name1, const char *name2);
 CONF_SECTION	*cf_section_sub_find(const CONF_SECTION *, const char *name);
 CONF_SECTION	*cf_section_sub_find_name2(const CONF_SECTION *, const char *name1, const char *name2);
 const char 	*cf_section_value_find(const CONF_SECTION *, const char *attr);
@@ -70,6 +77,7 @@ int cf_data_add(CONF_SECTION *, const char *, void *, void (*)(void *));
 
 const char *cf_pair_attr(CONF_PAIR *pair);
 const char *cf_pair_value(CONF_PAIR *pair);
+FR_TOKEN cf_pair_operator(CONF_PAIR *pair);
 VALUE_PAIR *cf_pairtovp(CONF_PAIR *pair);
 const char *cf_section_name1(const CONF_SECTION *);
 const char *cf_section_name2(const CONF_SECTION *);
@@ -85,6 +93,7 @@ int cf_pair_lineno(CONF_PAIR *pair);
 const char *cf_pair_filename(CONF_PAIR *pair);
 const char *cf_section_filename(CONF_SECTION *section);
 CONF_ITEM *cf_item_find_next(CONF_SECTION *section, CONF_ITEM *item);
+CONF_SECTION *cf_item_parent(CONF_ITEM *ci);
 int cf_item_is_section(CONF_ITEM *item);
 int cf_item_is_pair(CONF_ITEM *item);
 CONF_PAIR *cf_itemtopair(CONF_ITEM *item);
@@ -113,9 +122,18 @@ CONF_ITEM *cf_reference_item(const CONF_SECTION *parentcs,
 extern int cf_log_config;
 extern int cf_log_modules;
 
+extern int cf_pair2xml(FILE *fp, const CONF_PAIR *cp);
+extern int cf_section2xml(FILE *fp, const CONF_SECTION *cs);
+extern int cf_pair2file(FILE *fp, const CONF_PAIR *cp);
+extern int cf_section2file(FILE *fp, const CONF_SECTION *cs);
+
 /*
  *	Big magic.
  */
 int cf_section_migrate(CONF_SECTION *dst, CONF_SECTION *src);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* _CONFFILE_H */
