@@ -5,16 +5,23 @@
  * realms.h	Structures, prototypes and global variables
  *		for realms
  *
- * Version:	$Id$
+ * Version:	$Id: ffeaab07c0384b9de78e6b87daefbe9661bdc9e6 $
  *
  */
 
 #include <freeradius-devel/ident.h>
-RCSIDH(realms_h, "$Id$")
+RCSIDH(realms_h, "$Id: ffeaab07c0384b9de78e6b87daefbe9661bdc9e6 $")
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 #define HOME_TYPE_INVALID (0)
 #define HOME_TYPE_AUTH    (1)
 #define HOME_TYPE_ACCT    (2)
+#ifdef WITH_COA
+#define HOME_TYPE_COA     (3)
+#endif
 
 #define HOME_PING_CHECK_NONE		(0)
 #define HOME_PING_CHECK_STATUS_SERVER	(1)
@@ -45,9 +52,12 @@ typedef struct home_server {
 	struct timeval	when;
 
 	int		response_window;
+	int		no_response_fail;
 	int		max_outstanding; /* don't overload it */
 	int		currently_outstanding;
+	int		message_authenticator;
 
+	time_t		last_packet;
 	struct timeval	revive_time;
 	struct timeval	zombie_period_start;
 	int		zombie_period; /* unresponsive for T, mark it dead */
@@ -65,9 +75,17 @@ typedef struct home_server {
 
 	int		revive_interval; /* if it doesn't support pings */
 	CONF_SECTION	*cs;
+#ifdef WITH_COA
+	int			coa_irt;
+	int			coa_mrc;
+	int			coa_mrt;
+	int			coa_mrd;
+#endif
 #ifdef WITH_STATS
 	int		number;
 
+	fr_ipaddr_t	src_ipaddr; /* preferred source IP address */
+
 	fr_stats_t	stats;
 
 	fr_stats_ema_t  ema;
@@ -117,8 +135,17 @@ REALM *realm_find2(const char *name); /* ... with name taken from realm_find */
 
 home_server *home_server_ldb(const char *realmname, home_pool_t *pool, REQUEST *request);
 home_server *home_server_find(fr_ipaddr_t *ipaddr, int port);
+int	home_server_create_listeners(void);
+#ifdef WITH_COA
+home_server *home_server_byname(const char *name, int type);
+#endif
 #ifdef WITH_STATS
 home_server *home_server_bynumber(int number);
 #endif
+home_pool_t *home_pool_byname(const char *name, int type);
+
+#ifdef __cplusplus
+}
+#endif
 
 #endif /* REALMS_H */
