@@ -1,7 +1,7 @@
 /*
  * stats.c	Internal statistics handling.
  *
- * Version:	$Id$
+ * Version:	$Id: 6b7779ca3e826fa2a68e8b0e0b8dc22cef4d1829 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 6b7779ca3e826fa2a68e8b0e0b8dc22cef4d1829 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/rad_assert.h>
@@ -58,6 +58,17 @@ void request_stats_final(REQUEST *request)
 	    (request->listener->type != RAD_LISTEN_AUTH) &&
 	    (request->listener->type != RAD_LISTEN_ACCT)) return;
 
+	/* don't count statistic requests */
+	if (request->packet->code == PW_STATUS_SERVER)
+		return;
+
+#undef INC_AUTH
+#define INC_AUTH(_x) radius_auth_stats._x++;request->listener->stats._x++;if (request->client && request->client->auth) request->client->auth->_x++;
+
+
+#undef INC_ACCT
+#define INC_ACCT(_x) radius_acct_stats._x++;request->listener->stats._x++;if (request->client && request->client->acct) request->client->acct->_x++
+
 	/*
 	 *	Update the statistics.
 	 *
@@ -68,42 +79,23 @@ void request_stats_final(REQUEST *request)
 	 */
 	switch (request->reply->code) {
 	case PW_AUTHENTICATION_ACK:
-		radius_auth_stats.total_responses++;
-		radius_auth_stats.total_access_accepts++;
-		request->listener->stats.total_responses++;
-		request->listener->stats.total_access_accepts++;
-		if (request->client && request->client->auth) {
-			request->client->auth->total_access_accepts++;
-		}
+		INC_AUTH(total_responses);
+		INC_AUTH(total_access_accepts);
 		break;
 
 	case PW_AUTHENTICATION_REJECT:
-		radius_auth_stats.total_responses++;
-		radius_auth_stats.total_access_rejects++;
-		request->listener->stats.total_responses++;
-		request->listener->stats.total_access_rejects++;
-		if (request->client && request->client->auth) {
-			request->client->auth->total_access_rejects++;
-		}
+		INC_AUTH(total_responses);
+		INC_AUTH(total_access_rejects);
 		break;
 
 	case PW_ACCESS_CHALLENGE:
-		radius_auth_stats.total_responses++;
-		radius_auth_stats.total_access_challenges++;
-		request->listener->stats.total_responses++;
-		request->listener->stats.total_access_challenges++;
-		if (request->client && request->client->auth) {
-			request->client->auth->total_access_challenges++;
-		}
+		INC_AUTH(total_responses);
+		INC_AUTH(total_access_challenges);
 		break;
 
 #ifdef WITH_ACCOUNTING
 	case PW_ACCOUNTING_RESPONSE:
-		radius_acct_stats.total_responses++;
-		request->listener->stats.total_responses++;
-		if (request->client && request->client->acct) {
-			request->client->acct->total_responses++;
-		}
+		INC_ACCT(total_responses);
 		break;
 #endif
 
@@ -113,10 +105,16 @@ void request_stats_final(REQUEST *request)
 		 */
 	case 0:
 		if (request->packet->code == PW_AUTHENTICATION_REQUEST) {
-			radius_auth_stats.total_bad_authenticators++;
-			request->listener->stats.total_bad_authenticators++;
-			if (request->client && request->client->auth) {
-				request->client->auth->total_bad_authenticators++;
+			if (request->reply->offset == -2) {
+				INC_AUTH(total_bad_authenticators);
+			} else {
+				INC_AUTH(total_packets_dropped);
+			}
+		} else if (request->packet->code == PW_ACCOUNTING_REQUEST) {
+			if (request->reply->offset == -2) {
+				INC_ACCT(total_bad_authenticators);
+			} else {
+				INC_ACCT(total_packets_dropped);
 			}
 		}
 		break;
@@ -170,7 +168,7 @@ void request_stats_final(REQUEST *request)
 
 #ifdef WITH_ACCOUNTING
 	case PW_ACCOUNTING_RESPONSE:
-		radius_acct_stats.total_responses++;
+		proxy_acct_stats.total_responses++;
 		request->proxy_listener->stats.total_responses++;
 		request->home_server->stats.total_responses++;
 		break;
@@ -465,6 +463,8 @@ void request_stats_reply(REQUEST *request)
 			if (server_ip) {
 				pairadd(&request->reply->vps,
 					paircopyvp(server_ip));
+			}
+			if (server_port) {
 				pairadd(&request->reply->vps,
 					paircopyvp(server_port));
 			}
@@ -534,6 +534,7 @@ void request_stats_reply(REQUEST *request)
 #endif
 	}
 
+#ifdef WITH_PROXY
 	/*
 	 *	Home servers.
 	 */
@@ -623,6 +624,7 @@ void request_stats_reply(REQUEST *request)
 		}
 #endif
 	}
+#endif	/* WITH_PROXY */
 }
 
 void radius_stats_init(int flag)
