@@ -1,7 +1,7 @@
 /*
  * cb.c
  *
- * Version:     $Id$
+ * Version:     $Id: f7a0bebc488f19d3b0c3effbb89e46729cbb9002 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: f7a0bebc488f19d3b0c3effbb89e46729cbb9002 $")
 
 #include "eap_tls.h"
 
@@ -34,6 +34,7 @@ void cbtls_info(const SSL *s, int where, int ret)
 	int w;
 	EAP_HANDLER *handler = (EAP_HANDLER *)SSL_get_ex_data(s, 0);
 	REQUEST *request = NULL;
+	char buffer[1024];
 
 	if (handler) request = handler->request;
 
@@ -44,30 +45,47 @@ void cbtls_info(const SSL *s, int where, int ret)
 
 	state = SSL_state_string_long(s);
 	state = state ? state : "NULL";
+	buffer[0] = '\0';
 
 	if (where & SSL_CB_LOOP) {
-		RDEBUG2("%s: %s\n", str, state);
+		RDEBUG2("%s: %s", str, state);
 	} else if (where & SSL_CB_HANDSHAKE_START) {
-		RDEBUG2("%s: %s\n", str, state);
+		RDEBUG2("%s: %s", str, state);
 	} else if (where & SSL_CB_HANDSHAKE_DONE) {
-		RDEBUG2("%s: %s\n", str, state);
+		RDEBUG2("%s: %s", str, state);
 	} else if (where & SSL_CB_ALERT) {
 		str=(where & SSL_CB_READ)?"read":"write";
-		radlog(L_ERR,"TLS Alert %s:%s:%s\n", str,
-			SSL_alert_type_string_long(ret),
-			SSL_alert_desc_string_long(ret));
+
+		snprintf(buffer, sizeof(buffer), "TLS Alert %s:%s:%s",
+			 str,
+			 SSL_alert_type_string_long(ret),
+			 SSL_alert_desc_string_long(ret));
 	} else if (where & SSL_CB_EXIT) {
 		if (ret == 0) {
-			radlog(L_ERR, "%s:failed in %s\n", str, state);
+			snprintf(buffer, sizeof(buffer), "%s: failed in %s",
+				 str, state);
+
 		} else if (ret < 0) {
 			if (SSL_want_read(s)) {
 				RDEBUG2("%s: Need to read more data: %s",
 				       str, state);
 			} else {
-				radlog(L_ERR, "%s:error in %s\n", str, state);
+				snprintf(buffer, sizeof(buffer),
+					 "%s: error in %s", str, state);
 			}
 		}
 	}
+
+	if (buffer[0]) {
+		radlog(L_ERR, "%s", buffer);
+		
+		if (request) {
+			VALUE_PAIR *vp;
+			
+			vp = pairmake("Module-Failure-Message", buffer, T_OP_ADD);
+			if (vp) pairadd(&request->packet->vps, vp);
+		}
+	}
 }
 
 /*
@@ -118,6 +136,7 @@ int cbtls_password(char *buf,
  */
 int eaptls_handle_idx = -1;
 int eaptls_conf_idx = -1;
+int eaptls_store_idx = -1; /* OCSP Store */
 int eaptls_session_idx = -1;
 
 #endif /* !defined(NO_OPENSSL) */
