@@ -1,7 +1,7 @@
 /*
  * mem.c  Memory allocation, deallocation stuff.
  *
- * Version:     $Id$
+ * Version:     $Id: 5bfb8e61888a30cbd57e8e239c4eb4a4e6a4e7fd $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,11 +22,19 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 5bfb8e61888a30cbd57e8e239c4eb4a4e6a4e7fd $")
 
 #include <stdio.h>
 #include "rlm_eap.h"
 
+#ifdef HAVE_PTHREAD_H
+#define PTHREAD_MUTEX_LOCK pthread_mutex_lock
+#define PTHREAD_MUTEX_UNLOCK pthread_mutex_unlock
+#else
+#define PTHREAD_MUTEX_LOCK(_x)
+#define PTHREAD_MUTEX_UNLOCK(_x)
+#endif
+
 /*
  * Allocate a new EAP_PACKET
  */
@@ -112,20 +120,41 @@ void eap_ds_free(EAP_DS **eap_ds_p)
 /*
  * Allocate a new EAP_HANDLER
  */
-EAP_HANDLER *eap_handler_alloc(void)
+EAP_HANDLER *eap_handler_alloc(rlm_eap_t *inst)
 {
 	EAP_HANDLER	*handler;
 
 	handler = rad_malloc(sizeof(EAP_HANDLER));
 	memset(handler, 0, sizeof(EAP_HANDLER));
+
+	if (inst->handler_tree) {
+		PTHREAD_MUTEX_LOCK(&(inst->handler_mutex));
+		rbtree_insert(inst->handler_tree, handler);
+		PTHREAD_MUTEX_UNLOCK(&(inst->handler_mutex));
+	
+	}
 	return handler;
 }
 
-void eap_handler_free(EAP_HANDLER *handler)
+void eap_opaque_free(EAP_HANDLER *handler)
 {
 	if (!handler)
 		return;
 
+	eap_handler_free(handler->inst_holder, handler);
+}
+
+void eap_handler_free(rlm_eap_t *inst, EAP_HANDLER *handler)
+{
+	if (!handler)
+		return;
+
+	if (inst->handler_tree) {
+		PTHREAD_MUTEX_LOCK(&(inst->handler_mutex));
+		rbtree_deletebydata(inst->handler_tree, handler);
+		PTHREAD_MUTEX_UNLOCK(&(inst->handler_mutex));
+	}
+
 	if (handler->identity) {
 		free(handler->identity);
 		handler->identity = NULL;
@@ -144,14 +173,96 @@ void eap_handler_free(EAP_HANDLER *handler)
 	handler->opaque = NULL;
 	handler->free_opaque = NULL;
 
+	if (handler->certs) pairfree(&handler->certs);
+
 	free(handler);
 }
 
+
+typedef struct check_handler_t {
+	rlm_eap_t	*inst;
+	EAP_HANDLER	*handler;
+	int		trips;
+} check_handler_t;
+
+static void check_handler(void *data)
+{
+	int do_warning = FALSE;
+	uint8_t state[8];
+	check_handler_t *check = data;
+
+	if (!check) return;
+
+	if (!check->inst || !check->handler) {
+		free(check);
+		return;
+	}
+
+	if (!check->inst->handler_tree) goto done;
+
+	PTHREAD_MUTEX_LOCK(&(check->inst->handler_mutex));
+	if (!rbtree_finddata(check->inst->handler_tree, check->handler)) {
+		goto done;
+	}
+
+	/*
+	 *	The session has continued *after* this packet.
+	 *	Don't do a warning.
+	 */
+	if (check->handler->trips > check->trips) {
+		goto done;
+	}
+
+	/*
+	 *	No TLS means no warnings.
+	 */
+	if (!check->handler->tls) goto done;
+
+	/*
+	 *	If we're being deleted early, it's likely because we
+	 *	received a transmit from the client that re-uses the
+	 *	same RADIUS Id, which forces the current packet to be
+	 *	deleted.  In that case, ignore the error.
+	 */
+	if (time(NULL) < (check->handler->timestamp + 3)) goto done;
+
+	if (!check->handler->finished) {
+		do_warning = TRUE;
+		memcpy(state, check->handler->state, sizeof(state));
+	}
+
+done:
+	PTHREAD_MUTEX_UNLOCK(&(check->inst->handler_mutex));
+	free(check);
+
+	if (do_warning) {
+		DEBUG("WARNING: !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
+		DEBUG("WARNING: !! EAP session for state 0x%02x%02x%02x%02x%02x%02x%02x%02x did not finish!",
+		      state[0], state[1],
+		      state[2], state[3],
+		      state[4], state[5],
+		      state[6], state[7]);
+
+		DEBUG("WARNING: !! Please read http://wiki.freeradius.org/guide/Certificate_Compatibility");
+		DEBUG("WARNING: !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
+	}
+}
+
 void eaptype_free(EAP_TYPES *i)
 {
+	cf_section_parse_free(i->cs, i->type_data);
+
 	if (i->type->detach) (i->type->detach)(i->type_data);
 	i->type_data = NULL;
-	if (i->handle) lt_dlclose(i->handle);
+#ifndef NDEBUG
+	/*
+	 *	Don't dlclose() modules if we're doing memory
+	 *	debugging.  This removes the symbols needed by
+	 *	valgrind.
+	 */
+	if (!mainconfig.debug_memory)
+#endif
+	  if (i->handle) lt_dlclose(i->handle);
 	free(i);
 }
 
@@ -162,7 +273,7 @@ void eaplist_free(rlm_eap_t *inst)
 
        	for (node = inst->session_head; node != NULL; node = next) {
 		next = node->next;
-		eap_handler_free(node);
+		eap_handler_free(inst, node);
 	}
 
 	inst->session_head = inst->session_tail = NULL;
@@ -184,6 +295,7 @@ static uint32_t eap_rand(fr_randctx *ctx)
 	return num;
 }
 
+
 static EAP_HANDLER *eaplist_delete(rlm_eap_t *inst, EAP_HANDLER *handler)
 {
 	rbnode_t *node;
@@ -229,10 +341,15 @@ static void eaplist_expire(rlm_eap_t *inst, time_t timestamp)
 	 *	handlers to be deleted.
 	 *
 	 */
-	for (i = 0; i < 2; i++) {
+	for (i = 0; i < 3; i++) {
 		handler = inst->session_head;
-		if (handler &&
-		    ((timestamp - handler->timestamp) > inst->timer_limit)) {
+		if (!handler) break;
+
+		/*
+		 *	Expire entries from the start of the list.
+		 *	They should be the oldest ones.
+		 */
+		if ((timestamp - handler->timestamp) > inst->timer_limit) {
 			rbnode_t *node;
 			node = rbtree_find(inst->session_tree, handler);
 			rad_assert(node != NULL);
@@ -248,7 +365,7 @@ static void eaplist_expire(rlm_eap_t *inst, time_t timestamp)
 				inst->session_head = NULL;
 				inst->session_tail = NULL;
 			}
-			eap_handler_free(handler);
+			eap_handler_free(inst, handler);
 		}
 	}
 }
@@ -289,7 +406,7 @@ int eaplist_add(rlm_eap_t *inst, EAP_HANDLER *handler)
 	 *	Playing with a data structure shared among threads
 	 *	means that we need a lock, to avoid conflict.
 	 */
-	pthread_mutex_lock(&(inst->session_mutex));
+	PTHREAD_MUTEX_LOCK(&(inst->session_mutex));
 
 	/*
 	 *	If we have a DoS attack, discard new sessions.
@@ -338,6 +455,18 @@ int eaplist_add(rlm_eap_t *inst, EAP_HANDLER *handler)
 	 */
 	status = rbtree_insert(inst->session_tree, handler);
 
+	/*
+	 *	Catch Access-Challenge without response.
+	 */
+	if (inst->handler_tree) {
+		check_handler_t *check = rad_malloc(sizeof(*check));
+
+		check->inst = inst;
+		check->handler = handler;
+		check->trips = handler->trips;
+		request_data_add(request, inst, 0, check, check_handler);
+	}
+
 	if (status) {
 		EAP_HANDLER *prev;
 
@@ -364,7 +493,7 @@ int eaplist_add(rlm_eap_t *inst, EAP_HANDLER *handler)
 	 */
 	if (status > 0) handler->request = NULL;
 
-	pthread_mutex_unlock(&(inst->session_mutex));
+	PTHREAD_MUTEX_UNLOCK(&(inst->session_mutex));
 
 	if (status <= 0) {
 		pairfree(&state);
@@ -421,12 +550,12 @@ EAP_HANDLER *eaplist_find(rlm_eap_t *inst, REQUEST *request,
 	 *	Playing with a data structure shared among threads
 	 *	means that we need a lock, to avoid conflict.
 	 */
-	pthread_mutex_lock(&(inst->session_mutex));
+	PTHREAD_MUTEX_LOCK(&(inst->session_mutex));
 
 	eaplist_expire(inst, request->timestamp);
 
 	handler = eaplist_delete(inst, &myHandler);
-	pthread_mutex_unlock(&(inst->session_mutex));
+	PTHREAD_MUTEX_UNLOCK(&(inst->session_mutex));
 
 	/*
 	 *	Might not have been there.
@@ -438,7 +567,7 @@ EAP_HANDLER *eaplist_find(rlm_eap_t *inst, REQUEST *request,
 
 	if (handler->trips >= 50) {
 		RDEBUG2("More than 50 authentication packets for this EAP session.  Aborted.");
-		eap_handler_free(handler);
+		eap_handler_free(inst, handler);
 		return NULL;
 	}
 	handler->trips++;
