@@ -1,7 +1,7 @@
 /*
  * radeapclient.c	EAP specific radius packet debug tool.
  *
- * Version:	$Id$
+ * Version:	$Id: 7235969be34ca8a991c06e06d3fe56838e468587 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 7235969be34ca8a991c06e06d3fe56838e468587 $")
 
 #include <freeradius-devel/libradius.h>
 
@@ -83,12 +83,15 @@ static void NEVER_RETURNS usage(void)
 	fprintf(stderr, "  -f file     Read packets from file, not stdin.\n");
 	fprintf(stderr, "  -r retries  If timeout, retry sending the packet 'retries' times.\n");
 	fprintf(stderr, "  -t timeout  Wait 'timeout' seconds before retrying (may be a floating point number).\n");
+	fprintf(stderr, "  -h          Print usage help information.\n");
 	fprintf(stderr, "  -i id       Set request id to 'id'.  Values may be 0..255\n");
 	fprintf(stderr, "  -S file     read secret from file, not command line.\n");
 	fprintf(stderr, "  -q          Do not print anything out.\n");
 	fprintf(stderr, "  -s          Print out summary information of auth results.\n");
 	fprintf(stderr, "  -v          Show program version information.\n");
 	fprintf(stderr, "  -x          Debugging mode.\n");
+	fprintf(stderr, "  -4          Use IPv4 address of server\n");
+	fprintf(stderr, "  -6          Use IPv6 address of server.\n");
 
 	exit(1);
 }
@@ -144,6 +147,58 @@ static int getport(const char *name)
 	return ntohs(svp->s_port);
 }
 
+#define R_RECV (0)
+#define R_SENT (1)
+static void debug_packet(RADIUS_PACKET *packet, int direction)
+{
+	VALUE_PAIR *vp;
+	char buffer[1024];
+	const char *received, *from;
+	const fr_ipaddr_t *ip;
+	int port;
+
+	if (!packet) return;
+
+	if (direction == 0) {
+		received = "Received";
+		from = "from";	/* what else? */
+		ip = &packet->src_ipaddr;
+		port = packet->src_port;
+
+	} else {
+		received = "Sending";
+		from = "to";	/* hah! */
+		ip = &packet->dst_ipaddr;
+		port = packet->dst_port;
+	}
+
+	/*
+	 *	Client-specific debugging re-prints the input
+	 *	packet into the client log.
+	 *
+	 *	This really belongs in a utility library
+	 */
+	if ((packet->code > 0) && (packet->code < FR_MAX_PACKET_CODE)) {
+		printf("%s %s packet %s host %s port %d, id=%d, length=%d\n",
+		       received, fr_packet_codes[packet->code], from,
+		       inet_ntop(ip->af, &ip->ipaddr, buffer, sizeof(buffer)),
+		       port, packet->id, packet->data_len);
+	} else {
+		printf("%s packet %s host %s port %d code=%d, id=%d, length=%d\n",
+		       received, from,
+		       inet_ntop(ip->af, &ip->ipaddr, buffer, sizeof(buffer)),
+		       port,
+		       packet->code, packet->id, packet->data_len);
+	}
+
+	for (vp = packet->vps; vp != NULL; vp = vp->next) {
+		vp_prints(buffer, sizeof(buffer), vp);
+		printf("\t%s\n", buffer);
+	}
+	fflush(stdout);
+}
+
+
 static int send_packet(RADIUS_PACKET *req, RADIUS_PACKET **rep)
 {
 	int i;
@@ -152,6 +207,8 @@ static int send_packet(RADIUS_PACKET *req, RADIUS_PACKET **rep)
 	for (i = 0; i < retries; i++) {
 		fd_set		rdfdesc;
 
+		debug_packet(req, R_SENT);
+
 		rad_send(req, NULL, secret);
 
 		/* And wait for reply, timing out as necessary */
@@ -220,13 +277,11 @@ static int send_packet(RADIUS_PACKET *req, RADIUS_PACKET **rep)
 
 	/* libradius debug already prints out the value pairs for us */
 	if (!fr_debug_flag && do_output) {
-		printf("Received response ID %d, code %d, length = %d\n",
-				(*rep)->id, (*rep)->code, (*rep)->data_len);
-		vp_printlist(stdout, (*rep)->vps);
+		debug_packet(*rep, R_RECV);
 	}
 	if((*rep)->code == PW_AUTHENTICATION_ACK) {
 		totalapp++;
-	} else {
+	} else if ((*rep)->code == PW_AUTHENTICATION_REJECT) {
 		totaldeny++;
 	}
 
@@ -643,9 +698,6 @@ static int respond_eap_sim(RADIUS_PACKET *req,
 	 */
 	unmap_eapsim_types(req);
 
-	printf("<+++ EAP-sim decoded packet:\n");
-	vp_printlist(stdout, req->vps);
-
 	if((vp = pairfind(req->vps, ATTRIBUTE_EAP_SIM_SUBTYPE)) == NULL)
 	{
 		return 0;
@@ -813,9 +865,6 @@ static int sendrecv_eap(RADIUS_PACKET *rep)
  again:
 	rep->id++;
 
-	printf("\n+++> About to send encoded packet:\n");
-	vp_printlist(stdout, rep->vps);
-
 	/*
 	 * if there are EAP types, encode them into an EAP-Message
 	 *
@@ -887,8 +936,7 @@ static int sendrecv_eap(RADIUS_PACKET *rep)
 	/* okay got back the packet, go and decode the EAP-Message. */
 	unmap_eap_types(req);
 
-	printf("<+++ EAP decoded packet:\n");
-	vp_printlist(stdout, req->vps);
+	debug_packet(req, R_RECV);
 
 	/* now look for the code type. */
 	for (vp = req->vps; vp != NULL; vp = vpnext) {
@@ -929,15 +977,22 @@ int main(int argc, char **argv)
 	FILE *fp;
 	int count = 1;
 	int id;
+	int force_af = AF_UNSPEC;
 
 	id = ((int)getpid() & 0xff);
 	fr_debug_flag = 0;
 
 	radlog_dest = RADLOG_STDERR;
 
-	while ((c = getopt(argc, argv, "c:d:f:hi:qst:r:S:xXv")) != EOF)
+	while ((c = getopt(argc, argv, "46c:d:f:hi:qst:r:S:xXv")) != EOF)
 	{
 		switch(c) {
+		case '4':
+			force_af = AF_INET;
+			break;
+		case '6':
+			force_af = AF_INET6;
+			break;
 		case 'c':
 			if (!isdigit((int) *optarg))
 				usage();
@@ -987,7 +1042,7 @@ int main(int argc, char **argv)
 			timeout = atof(optarg);
 			break;
 		case 'v':
-			printf("radclient: $Id$ built on " __DATE__ " at " __TIME__ "\n");
+			printf("radclient: $Id: 7235969be34ca8a991c06e06d3fe56838e468587 $ built on " __DATE__ " at " __TIME__ "\n");
 			exit(0);
 			break;
                case 'S':
@@ -1060,11 +1115,45 @@ int main(int argc, char **argv)
 	req->id = id;
 
 	/*
-	 *	Strip port from hostname if needed.
+	 *	Resolve hostname.
 	 */
-	if ((p = strchr(argv[1], ':')) != NULL) {
-		*p++ = 0;
-		port = atoi(p);
+	if (force_af == AF_UNSPEC) force_af = AF_INET;
+	req->dst_ipaddr.af = force_af;
+	if (strcmp(argv[1], "-") != 0) {
+		const char *hostname = argv[1];
+		const char *portname = argv[1];
+		char buffer[256];
+
+		if (*argv[1] == '[') { /* IPv6 URL encoded */
+			p = strchr(argv[1], ']');
+			if ((size_t) (p - argv[1]) >= sizeof(buffer)) {
+				usage();
+			}
+
+			memcpy(buffer, argv[1] + 1, p - argv[1] - 1);
+			buffer[p - argv[1] - 1] = '\0';
+
+			hostname = buffer;
+			portname = p + 1;
+
+		}
+		p = strchr(portname, ':');
+		if (p && (strchr(p + 1, ':') == NULL)) {
+			*p = '\0';
+			portname = p + 1;
+		} else {
+			portname = NULL;
+		}
+
+		if (ip_hton(hostname, force_af, &req->dst_ipaddr) < 0) {
+			fprintf(stderr, "radclient: Failed to find IP address for host %s: %s\n", hostname, strerror(errno));
+			exit(1);
+		}
+
+		/*
+		 *	Strip port from hostname if needed.
+		 */
+		if (portname) port = atoi(portname);
 	}
 
 	/*
@@ -1097,15 +1186,7 @@ int main(int argc, char **argv)
 	} else {
 		usage();
 	}
-
-	/*
-	 *	Resolve hostname.
-	 */
 	req->dst_port = port;
-	if (ip_hton(argv[1], AF_INET, &req->dst_ipaddr) < 0) {
-		fprintf(stderr, "radclient: Failed to find IP address for host %s\n", argv[1]);
-		exit(1);
-	}
 
 	/*
 	 *	Add the secret.
@@ -1392,13 +1473,18 @@ main(int argc, char *argv[])
 			break;
 		}
 
-		printf("\nRead:\n");
-		vp_printlist(stdout, req->vps);
+		if (fr_debug_flag > 1) {
+			printf("\nRead:\n");
+			vp_printlist(stdout, req->vps);
+		}
 
 		map_eapsim_types(req);
 		map_eap_types(req);
-		printf("Mapped to:\n");
-		vp_printlist(stdout, req->vps);
+
+		if (fr_debug_flag > 1) {
+			printf("Mapped to:\n");
+			vp_printlist(stdout, req->vps);
+		}
 
 		/* find the EAP-Message, copy it to req2 */
 		vp = paircopy2(req->vps, PW_EAP_MESSAGE);
@@ -1411,8 +1497,10 @@ main(int argc, char *argv[])
 		unmap_eap_types(req2);
 		unmap_eapsim_types(req2);
 
-		printf("Unmapped to:\n");
-		vp_printlist(stdout, req2->vps);
+		if (fr_debug_flag > 1) {
+			printf("Unmapped to:\n");
+			vp_printlist(stdout, req2->vps);
+		}
 
 		vp = pairfind(req2->vps,
 			      ATTRIBUTE_EAP_SIM_BASE+PW_EAP_SIM_MAC);
